exclude: true

```{r setup, message=FALSE, warning=FALSE, include=FALSE}
options(
  htmltools.dir.version = FALSE, # for blogdown
  width=80
)

library(tidyverse)
library(repurrrsive)
```

---
class: middle
count: false

# Functional Programming

---

## Functions as objects

We have mentioned in passing that functions in R are treated as 1st class objects (like vectors) 

.pull-left[
```{r}
f = function(x) {
  x*x
}

f(2)

g = f

g(2)
```
]

--

.pull-right[
```{r error=TRUE}
l = list(f = f, g = g)

l$f(3)
l[[2]](4)
l[1](3)
```
]

---

## Functions as arguments

We can pass in functions as arguments to other functions,

```{r}
do_calc = function(v, func) {
  func(v)
}
```

--

```{r}
do_calc(1:3, sum)
do_calc(1:3, mean)
do_calc(1:3, sd)
```


---

## Anonymous functions

These are short functions that are created without ever assigning a name,

```{r}
function(x) {x+1}

(function(y) {y-1})(10)
```

--

this can be particularly helpful for implementing certain types of tasks,

```{r}
integrate(function(x) x, 0, 1)

integrate(function(x) x^2-2*x+1, 0, 1)

integrate(function(x) sin(x)^2+cos(x)^2, 0, pi)
```


---
class: middle
count: false

# apply (base R)

---

## Apply functions

The apply functions are a collection of tools for functional programming in base R, they are variations of the `map` function found in many other languages and apply a function over the elements of the input.

```{r, eval=FALSE}
??base::apply
---

## 
## Help files with alias or concept or title matching ‘apply’ using fuzzy
## matching:
## 
## base::apply             Apply Functions Over Array Margins
## base::.subset           Internal Objects in Package 'base'
## base::by                Apply a Function to a Data Frame Split by Factors
## base::eapply            Apply a Function Over Values in an Environment
## base::lapply            Apply a Function over a List or Vector
## base::mapply            Apply a Function to Multiple List or Vector Arguments
## base::rapply            Recursively Apply a Function to a List
## base::tapply            Apply a Function Over a Ragged Array
```

---

## lapply

Usage: `lapply(X, FUN, ...)`

`lapply` returns a list of the same length as `X`, each element of which is the result of applying `FUN` to the corresponding element of `X`.

<br/>

.pull-left[
```{r}
lapply(1:8, sqrt) %>% str()
```
]

.pull-right[
```{r}
lapply(1:8, function(x) (x+1)^2) %>% str()
```
]

---

```{r}
lapply(1:8, function(x, pow) x^pow, pow=3) %>% str()
```

--

```{r}
lapply(1:8, function(x, pow) x^pow, x=2) %>% str()
```

---

## sapply

Usage: `sapply(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE)`

`sapply` is a *user-friendly* version and wrapper of `lapply`, it is a *simplifying* version of lapply. Whenever possible it will return a vector, matrix, or an array.

<br/>

```{r}
sapply(1:8, sqrt)
sapply(1:8, function(x) (x+1)^2)
sapply(1:8, function(x) c(x, x^2, x^3, x^4))
```

---

What happens if the returned lengths don't match?

.pull-left[
```{r}
sapply(1:6, seq)
```
]

--

.pull-right[
```{r}
lapply(1:6, seq)
```
]

---

What happens if the types don't match?

```{r}
l = list(a = 1:3, b = 4:6, c = 7:9, d = list(10, 11, "A"))
```

```{r}
sapply(l, function(x) x[1])
```

--

```{r}
sapply(l, function(x) x[[1]])
```

--

```{r}
sapply(l, function(x) x[[3]])
```

---

## `*`apply and data frames

We can use these functions with data frames, the key is to remember that a data frame is just a fancy list.

```{r}
df = data.frame(
  a = 1:6, 
  b = letters[1:6], 
  c = c(TRUE,FALSE)
)
```

```{r}
lapply(df, class) %>% str()
sapply(df, class)
```

---

## A more useful example

Recall Exercise 2 from Lecture 5 where we converted `-999` to `NA`s.

```{r}
d = data.frame(
  patient_id = c(1, 2, 3, 4, 5),
  age = c(32, 27, 56, 19, 65),
  bp = c(110, 100, 125, -999, -999),
  o2 = c(97, 95, -999, -999, 99)
)
```

--

.pull-left[
```{r}
fix_missing = function(x) {
  x[x == -999] = NA
  x
}
lapply(d, fix_missing)
```
]

--

.pull-right[
```{r}
lapply(d, fix_missing) %>%
  as.data.frame()
```
]

---

dplyr is also a viable option here, 

```{r}
d %>%
  mutate_if(is.numeric, fix_missing)
```


---

## other less common apply functions

* `apply()` - applies a function over the rows or columns of a data frame, matrix or array

* `vapply()` - is similar to `sapply`, but has a enforced return type and size

* `mapply()` -  like `sapply` but will iterate over multiple vectors at the same time.

* `rapply()` - a recursive version of `lapply`, behavior depends largely on the `how` argument

* `eapply()` -  apply a function over an environment.

---
class: middle

.center[
```{r echo=FALSE, out.width="45%"}
knitr::include_graphics('imgs/hex-purrr.png')
```
]

---

## Map functions

Basic functions for looping over objects and returning a value (of a specific type) - replacement for `lapply`/`sapply`/`vapply`.

* `map()` - returns a list.

* `map_lgl()` - returns a logical vector.

* `map_int()` - returns a integer vector.

* `map_dbl()` - returns a double vector.

* `map_chr()` - returns a character vector.

* `map_dfr()` - returns a data frame by row binding.

* `map_dfc()` - returns a data frame by column binding.

* `walk()` - returns nothing, function used exclusively for its side effects


---

## Type Consistency

R is a weakly / dynamically typed language which means there is no simple way to define a function which enforces the argument or return types. This flexibility can be useful at times, but often it makes it hard to reason about your code and requires more verbose code to handle edge cases.

```{r}
x = list(rnorm(1e3), rnorm(1e3), rnorm(1e3))
```

--

```{r}
map_dbl(x, mean)
```

--

```{r}
map_chr(x, mean)
```

--

```{r error=TRUE}
map_int(x, mean)
```

--

```{r}
map(x, mean) %>% str()
```

---

## Working with Data Frames

`map_dfr` and `map_dfc` are particularly useful when working with and/or creating data frames. 

Take for example the Lecture 5 Exercise 2 example from above,

.pull-left[
```{r}
d = data.frame(
  patient_id = c(1, 2, 3, 4, 5),
  age = c(32, 27, 56, 19, 65),
  bp = c(110, 100, 125, -999, -999),
  o2 = c(97, 95, -999, -999, 99)
)
```
]

.pull-right[
```{r}
fix_missing = function(x) {
  x[x == -999] = NA
  x
}
```
]

<br/>

```{r}
purrr::map_dfc(d, fix_missing)
```

---

```{r}
map_dfr(head(sw_people, 10), function(x) x[1:5])
```

--

<br/>

```{r error=TRUE}
map_dfr(head(sw_people, 10), function(x) x)
```

---

## Shortcut - Anonymous Functions

purrr lets us write anonymous functions using one sided formulas where the argument is given by `.` or `.x` for `map` and related functions.

```{r}
map_dbl(1:5, function(x) x/(x+1))
```

--

<br/>
```{r}
map_dbl(1:5, ~ ./(.+1))
```

--

<br/>
```{r}
map_dbl(1:5, ~ .x/(.x+1))
```

--

<br/>

Generally, the latter option is preferred to avoid confusion with magrittr.

---

## Shortcut - Anonymous Functions - `map2`

Functions with the `map2` prefix work the same as the `map` functions but they iterate over two objects instead of one. Arguments in an anonymous function are given by `.x` and `.y` (or `..1` and `..2`) respectively.

```{r}
map2_dbl(1:5, 1:5, function(x,y) x / (y+1))

map2_dbl(1:5, 1:5, ~ .x/(.y+1))

map2_dbl(1:5, 1:5, ~ ..1/(..2+1))
```

--

```{r}
map2_chr(LETTERS[1:5], letters[1:5], paste0)
```

---

## Shortcut - Lookups

Very often we want to extract only certain (named) values from a list, `purrr` provides a shortcut for this operation when you provide either a character or numeric value instead of a function to apply.

--

```{r}
purrr::map_chr(head(sw_people), "name")
```

--

```{r}
purrr::map_chr(head(sw_people), 1)
```

--

```{r}
purrr::map_chr(head(sw_people), list("films", 1))
```

---

## Length coercion?

```{r error = TRUE}
purrr::map_chr(head(sw_people), list("starships", 1))
```
--

<br/>
```{r}
sw_people[[2]]$name
sw_people[[2]]$starships
```

--

<br/>
```{r error = TRUE}
purrr::map_chr(head(sw_people), list("starships", 1), .default = NA)
```

---

## list columns

.pull-left[
```{r}
(chars = tibble(
  name = purrr::map_chr(sw_people, "name"),
  starships = purrr::map(sw_people, "starships")
))
```
]

.pull-right[
```{r}
chars %>%
  mutate(
    n_starships = map_int(starships, length)
  )
```
]

---
class: middle, center

## Demo - list columns and approximating pi



---
class: middle
count: false

# Acknowledgments

---

## Acknowledgments

Above materials are derived in part from the following sources:

* Hadley Wickham - [Adv-R Functionals](http://adv-r.had.co.nz/Functionals.html)

* Neil Saunders - [A brief introduction to "apply" in R](http://nsaunders.wordpress.com/2010/08/20/a-brief-introduction-to-apply-in-r/)

* Jenny Bryan - [Purrr Tutorial](https://jennybc.github.io/purrr-tutorial/)

* [R Language Definition](http://stat.ethz.ch/R-manual/R-devel/doc/manual/R-lang.html)


